// TLM_GPU header file Cuda TLM project

/*
    Copyright 2010 Computational Electromagnetics Research Laboratory (CERL)
    University of Victoria, Victoria, B.C. Canada
    http://www.cerl.ece.uvic.ca
    http://www.cerl.ece.uvic.ca/Open_Source_CEM.htm
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <oclUtils.h>
#include "timer.h"
#include <sys/stat.h>

#define VEL_LIGHT  299792458  // m/sec
#define PI 3.1415926535897
#define BLOCK_SIZE_X 64
#define BLOCK_SIZE_Y 1
#define BLOCK_SIZE_Z 1
#define MAX_GPU_COUNT 8


#pragma region MACROS_&_DEFINITIONS
	#define USE_MULTIPLE_GPUS	false
	// Defines for error handling
	#define STRINGIFY(x) #x
	#define TOSTRING(x) STRINGIFY(x)
	#define AT __FILE__ ":" TOSTRING(__LINE__)
	#define	OCL_ERROR_CHECK(clErrorNum)	checkErrorNumber(clErrorNum,AT)
	// Maximum devices to use for computation
	#define MAX_DEVICE_USE	8

	#define NODE_MEM_SIZE	12*sizeof(float)
#pragma endregion







namespace GPU_Space{
	class excitation; 
	class sampling_probe; 

	struct iterationSetupType{
		int meshDim_X,meshDim_Y,meshDim_Z; // overall mesh X,Y,Z dimensions
		int thread_BLK_X,thread_BLK_Y,thread_BLK_Z; // thread block dimensions (in threads)likely set to 64,1,1 for this specific implementation
		int blocksPerXGrid;   // fn(MeshX/thread_BLK_X)
		int blocksPerYGrid; // fn(MeshY/thread_BLK_Y)
		int nodes_per_Zslice;  // Obsolete in this implementation: CUDA does not support Z direction block processing so slices of the mesh were required
		int slices_Per_Zdim;   // Obsolete: used with nodes_per_Zslice
		//dim3 threads;   // threads and grid are passed to kernel as 'cuda_fn<<<grid,threads>>>  
		//dim3 grid;		//		where threads =(thread_BLK_X,thread_BLK_Y,thread_BLK_Z) and grid = (blocksPerXGrid,blocksPerYGrid,1)
		int threads[3];
		int grid[3];

	};

	class GPU{

		
	
	public: 
		// default constructor
		GPU(); // blank slate: paramters must be created later 
		GPU(char* filename, float scaling_mesh); // read in structure definition file
		// constructor with only mesh dimensions, therefore a blank mesh is instantiated. 
		GPU(int mesh_X, int mesh_Y, int mesh_Z, float resolution); 
		// constructor with mesh parameters: Pass in existing mesh data from CPU to be converted to GPU format
		GPU(float* mesh_data_parent, int mesh_X, int mesh_Y, int mesh_Z , float resolution) ;
		// constructor with mesh parameters + GPU device index (if available) 
		GPU(float* mesh_data_parent, int mesh_X, int mesh_Y, int mesh_Z , float resolution, int device) ;
		// deconstructor: free memory for both CPU and GPU data
		~GPU();

		
		
		///////////////////////
		// OCL PUBLIC VARIABLE Declaration
		//

		/*cl_context cxGPUContext;
		cl_uint ciDeviceCount;
		cl_kernel kernel_cmd_01[MAX_GPU_COUNT];
		cl_kernel kernel_cmd_02[MAX_GPU_COUNT];
		cl_command_queue commandQueue[MAX_GPU_COUNT];
		cl_program cpProgram;
		size_t localWorkSize[3];
		size_t globalWorkSize[3];
		cl_int errCode;
		cl_int ciErrNum ;
		int memSize_EH_field; 
		int memSize_Coef; 
		cl_event GPUDone[MAX_GPU_COUNT];
		cl_event GPUExecution[MAX_GPU_COUNT];*/

		/********************************* STATIC MEMBERS **********************************************/
		static char* source;								// Computation Object Source container
		static size_t programLength;						// Computation Object source code length
		static const char * binaryFile;						// name of binary file
		static const char * headerSource;					// name of kernel header file
		static const char * kernelSource;					// name of kernel source file
		static bool clInitialized;							// boolean for determining if the context has been initialized

		/********************************* NON-STATIC MEMBERS ******************************************/
		cl_context clContext;								// Computation Object context
		cl_program clProgram;								// Computation Object program
		
		cl_kernel kernel_v2_TLM_3D_Iterate_scatter[MAX_DEVICE_USE];			
		cl_kernel kernel_v2_TLM_3D_Iterate_scatter_interchange[MAX_DEVICE_USE];			
		cl_kernel kernel_v2_TLM_3D_Boundary_2[MAX_DEVICE_USE];			// TLM Scattering kernel object
		cl_kernel kernel_v2_TLM_3D_Plane_HalfSine_Excitation[MAX_DEVICE_USE];			
		cl_kernel kernel_v2_TLM_3D_Sample[MAX_DEVICE_USE];			


		cl_command_queue clCommandQueue[MAX_DEVICE_USE];	// Computation Object command queue
		cl_int clErrNum;									// Computation Object error number
		cl_uint clDeviceCount;								// Number of devices in Computation Object context
		cl_device_id *clDevices;							// pointer to Computation Object devices
		
		// Mesh memory allocation

		cl_mem mesh_device[MAX_DEVICE_USE];					// CL_MEM buffer object for device mesh
		//float* mesh_host;									// Pointer to host mesh data
		//UINT meshNodeSize;

		void oclInitialize();								// Initialize the OpenCL computation environment
		void Mesh_Device_to_host_memcopy(void);				// memCopy operation copies data from device to host
		void Mesh_Host_to_device_memcopy(void);				// memCopy operation copies data from host to device

		void createContext();								// Creates openCL computation context
		void getDevices();									// Gets the available openCL devices
		bool createProgram();								// Creates program object using loaded source code
		bool buildProgram();								// Builds program binaries with openCL compiler
		void createCommandQueues();							// Creates command queues for each available device
		void createKernels();								// Creates kernel objects for each available device
		void appendToSource(const char* path);				// Appends a file to the source code variable
		void createProgramFromSource();						// Creates a program from source
		void createProgramFromBinary();						// Creates a program from existing binary file
		void dumpPTX(const char * fileName);				// Dumps the binaries of the program to a file
		const char* getCLErrorDescription(int errorCode);	// Returns textual description of openCL error codes
		
		// This function should never be called directly.  Use the Macro OCL_ERROR_CHECK
		bool checkErrorNumber(const cl_int errorNumber,const char* location);	// Handles errors that occur in the oclTLM object

		//
		///////////////////////
		
		/////////////////////////////
		// 
		int iterations_MaxNum; // Maximum number of iterations set to execute
		int iteration_index; // current iteration number in iteration for next loop
		float resolution;// (m) between nodes
		int boundaries_Num;  // max number of boundaries defined
		float scaling_factor;		
		//
		////////////////////////

		
		////////////////////////
		// OCL calls
		void oclContext_CmndQueue_Create();
		void oclMeshMem_GPU_create();
		void oclKernel_Read_Compile();
		void oclKernel_TLM_Functions_create();
		void oclKernel_TLM_ITERATE();
		void oclMesh_Host_to_GPU_copy();
		void oclMesh_GPU_to_Host_copy();
		//
		///////////////////////

		/////////////////////
		// GPU Device Methods
		
		// allocates GPU global memory with mesh dimensions
		void cuda_Mesh_Allocate(float* mesh_host,int meshDim_X,int meshDim_Y,int meshDim_Z);
		
		// copies CPU mesh to GPU mesh
		void cuda_mesh_CPU_to_GPU_copy(); // wording: copy_to_GPU
		
		// queries the number of available GPU in cluster
		int  cuda_NumOfDevices_get();
		//
		//////////////////////
		
		

		/////////////////
		// MESH METHODS
		//
		// used to return the address (in an array of floats of a Mesh) of the 
		//    voltage arm in a node at x,y,z position. voltage arm index between (1 to 12) (do not use zero)
		int address(int voltage_arm, int x, int y, int z);
		
		// check if mesh is alread instantiated. If so then delete 
		// instantiate new GPU mesh (only one mesh allowed at a time)
		void Mesh_Host_create(int x, int y, int z); 
		
		// copies values from the Client's mesh to the GPU mesh
		// TBD: still addresses in old version of code, must convert but will wait till an interface with Mefisto is developed
		void Mesh_Client2GPU_set(float* mesh_host, int mesh_X, int mesh_Y, int mesh_Z );
		
		// Copies the GPU mesh data to the CPU (host)
		//void Mesh_Device_to_host_memcopy();

		// Get Mesh Data from GPU  by populating the CPU data
		// TBD: using old conversion so will wait till Mefisto interface developed
		void Mesh_GPU2Client_get(float* mesh_host, int* mesh_X, int* mesh_Y, int* mesh_Z );

		// Retrieves Mesh dimensions x,y,z
		void Mesh_Dimensions_get(int* x, int* y, int* z); 

		// Get a voltage arm value from Node at x,y,z from GPU_Mesh 
		// if armID ==0 then returns an average of the 12 voltage arms. 
		float Mesh_Node_get(int x, int y, int z, int armID );
		
		// TBD: when Mefisto interface completed
		//Array2D  float??? Mesh_data_plane_get(int xlow, int xhi, int ylow, int yhi, int zlow, int zhi);
		//Array3D float??? Mesh_data_cube_get(int xlow, int xhi, int ylow, int yhi, int zlow, int zhi);

		// Sets a voltage arm value at Node at x,y,z int GPU_Mesh 
		// if armID ==0 then sets all 12 voltage arms. 
		void Mesh_Node_set(float node_value, int x, int y, int z, int armID=0);
		
		// Pass a plane of nodes in an array of data, "mesh_data" and set a plane in the GPU_Mesh. 
		// TBD: future implementation
		void Mesh_Node_plane_set(float* mesh_host, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi, int armID=0);		
		
		// Set a 3D volume in GPU_Mesh with values passed into method. (Note: all 12 arms of nodes in defined volume will be set
		// TBD: future implementation
		void Mesh_Node_cube_set(float* mesh_host, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi);
		
		// Reads a csv file from 'filename' that contains: mesh dimensions, resolution, and boundaries
		// See "STRUCUTRE_DEFINITIONS.csv" for an example
		void read_Structure_Definition_file(char* filename);

		void strucuture_scaling(float scale_strucutre);
		
		// Sets constants in the GPU to be used for all mesh realted calculations (iteration, boundary, excitation...)
		// sets the following:  meshDim_X,meshDim_Y,meshDim_Z,GPU_BLOCK_X,GPU_BLOCK_Y,GPU_BLOCK_Z 
		void GPU::ocl_Mesh_Block_constants_set();
		//
		/////////////////


		/////////////////
		// BOUNDARY METHODS
		
		// Adds a boundary: TBD: not implemented at this time
		void bndry_add(float ref_coef, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi); // adds one at at time to list
		
		// Retrieves an existing boundary: TBD: not implemented yet
		void bndry_get(int bdry_index, float* ref_coef, int* xlow, int* xhi, int* ylow, int* yhi, int* zlow, int* zhi);
		
		// Sets an existing boundary with new data: TBD: not implemented yet
		void bndry_set(int bdry_index, float ref_coef, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi);
		
		//void bndry_seed_boundaries(); // OBSOLETE
		//void bndry_Embed_Boundary_Data(int x, int y , int z,int bndry, int ref_coef_index); // OBSOLETE
		
		// Sets GPU constants with an array of reflection coeficients to be used as a lookup for boundary routine
		void cuda_reflect_coef_constants_set();
		
		// Executes the boundary kernel in the GPU on 'boundaryID'
		void GPU::cuda_boundary_execute(int boundaryID);
		//////////////////
		
		/////////////////
		// ITERATION METHODS
		// populates the  'iterationSetup' structure with information required to execute kernels
		void cuda_iteration_setup();  
		void iterate();  // Executes a kernel iteration (SCATTERING) 
		//////////////////


		///////////////
		// SAMPLING PROBES
		sampling_probe* sampling_obj[100]; // an array of sampling class pointers : sampling class used to sample a point in a mesh
		int sample_Probe_add(int sample_maxnum, int polerize_dir, int x, int y, int z); // adds one at at time to list
		void sample_Probe_get(int probe_index, float* data, int* sample_maxnumint, int armID=0); // TBD: not implemented yet
		void GPU::sample_execute(int sample_ID); // executes the cuda sampling kernel and adds sample to the sampling array in sampling class
		void GPU::sample_gpu_sampleArray_Allocate(int sample_ID); // Calls cuda code to allocate sample array memory in GPU
		float* GPU::sample_gpu_samples_get(int sample_ID,int num_of_samples); // calls cuda code to retrieve sample data in GPU and copy to this object in CPU
		void GPU::sample_save_to_file(char* filename, int sample_ID); // Save the sample data to file (comma separated format)
		///////////////
		
		////////////////
		// EXCITATION
		// General Notes: three stages: 1) Create an excitation strucuture ('excitation_CreateStructure')  2) generate a waveform ('excitation_generate_SineGaussian_array') 3) execute the excitation 'excitation_execute_HalfSine'

		excitation* excite_obj[100];// an array of excitation class pointers: excitation class used to excite a plane in mesh structure
		
		// Creates a plane excitation structure: prop_dir = propatation direction (1=x direction, 2= y direction, 3= z direction)
		//   polz_dir = polerization direction (1=x dir, 2=y dir, 3=z dir)
		int excitation_CreateStructure(int prop_dir, int polz_dir, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi);
		
		// After creating an excitation structure then can generate a gaussian excitation waveform profile (in time steps)
		// freq (Hz), magnitude (Volts), Width (of gaussian waveform in time steps), peak_position (the peak pos of gaussian wave in time-steps, excite_array_size = size of array
		void excitation_generate_SineGaussian_array(int excite_obj_ID,float freq, float magnitude, float width, float peak_position,  int excite_array_size);
		
		// Execute an impulse excitation at an x,y,z coord (all 12 voltages) note: does not need a stucture nor a waveform, can be stand alone call
		void GPU::excitation_execute_Impulse(int x, int y, int z, float excitation);
		
		// Execute a Half Sine impulse. Needs a structure defined, but does not need a waveform defined. 
		void GPU::excitation_execute_HalfSine_Impulse(int excitation_ID);
		
		// Execute a Half Sine waveform. Needs a structure, and a waveform defined. And upon subsequent calls in an iteration loop 
		//   this will automatically advance the waveform timestep array index
		void GPU::excitation_execute_HalfSine(int excitation_ID);
		
		// Saves the excitation wavform profile
		void excitation_Save(int excite_obj_ID,char* filename);
		//
		////////////////
		

		////////////////////
		// Misc Methods
		void display_percent_done(); // displays to the command window periodic status info. (% complete, iteration#, Rate of execution)
		bool iteration_timer_display_ENABLE;  //enables display_percent_done()
		timer iteration_timer; // used by iteration timer to time exectution 
		double iteration_timer_lastTime;  // temp storage used by iteration timer
		int iteration_previously_saved_idx;  // temp storage used by iteration timer
		double MegaNodes_per_sec_running_avg_savd;// temp storage used by iteration timer
		void Output_Node_Layer_XY_Plane(char* filename, int z_layer, int voltage_arm ); // save to file a slice of node/voltage-arm data from a mesh int the X-Y plane (if voltage arm = 0 then node average saved)
		//
		///////////////////

	
		int meshDim_X,meshDim_Y,meshDim_Z;  // Mesh dimensions
	private: 
		//int device_number; 
		iterationSetupType iterationSetup;
		
		int GPU_BLOCK_X,GPU_BLOCK_Y,GPU_BLOCK_Z;  // defines the thread block dimensions. for this implementation GPU_BLOCK_X = 64, GPU_BLOCK_Y=1,GPU_BLOCK_Z=1
		int meshNodeSize ; // =meshDim_X*meshDim_Y*meshDim_Z
		float delta_time;  // Timpstep (s)
		float* mesh_client;  // pointer to the client's mesh data. Will implement more when interfacing with Mefisto
		float*  mesh_host;   // pointer to CPU host mesh data. Note: Both CPU Host and GPU mesh are required and are organized the same way. 
		float* GPU_Device_Mesh; // pointer to GPU device global mem. (address in GPU only)
		void CPUmesh_to_GPUmesh_convert(); // TBD: not implemented yet
		void GPUmesh_to_CPUmesh_convert(); // TBD: not implemented yet

		// BOUNDARY 
		int bndry_maxNum;  // Number of boundaries defined for this mesh
		struct BoundaryType { // each boundary defined by the following parameters
			int ref_coef_index; // 0-15  location in reflection index array 
			float reflection_coef;  // Not required since relction_coef array is sent to GPU as constant array and 'ref_coef_index' used to look up
			int Xmin; 
			int Xmax; 
			int Ymin;
			int Ymax; 
			int Zmin; 
			int Zmax;
		};
		float bndry_reflection_coef_list[16]; // a list of all possible boundary coeficients to be sent to GPU as constant array and then boundary kernel uses this as a lookup table
		
		BoundaryType* boundaries; // to be instantiated with "boundaries = NEW BoundaryType[bndry_maxNum]"  or malloc when the number of boundaries are determined
		
		
		void bndry_seed(int bndry_maxNum, BoundaryType* boundaries); //OBSOLETE: but may be usefull later ( seed the mesh with all boundaries)
	




	
	
	}; // end of Class GPU
/////////////////////////////////	
// Samples a node in the x,y or z polerization directions (i.e. will average 4 voltage arms that point in the poerization direction chosen)
// Samples are stored in a sample array of size 'sample_numMax'
class sampling_probe{
	public: 
		static int sampling_obj_count; // number of Sampling objects instantiated
		sampling_probe();  
		sampling_probe(int sample_maxnum, int polerize_dir, int x, int y, int z);
		
		float* samples;// to be instantiated with "samples = NEW float[sample_numMax]"  
		cl_mem gpu_samples; // pointer to GPU global data for samples
		int sample_index; // increments after each sample execution
		int sample_numMax; 	// max number of samples in array
		int sample_polarize_dir;  // 1= x polerization 2 = y polerization, 3 = z polerization
		int x;
		int y; 
		int z;
		
	private :

	
	}; // endo fo Sampling_probe class
////////////////////////////////


//////////////////////////////////

	class excitation {
		
	public: 
		static int excitation_obj_count;
		 
		//enum Excitation_type { NONE=0, POINT=1, HALF_SINE=2, GAUSSIAN_PULSE=3 };
		
		//ALL = 0, XposYpol = 11, XnegYpol = 3, XposZpol = 10, XnegZpol = 6, 
		//		   YposXpol = 12, YnegXpol = 1, YposZpol =  7, YnegZpol = 5, 
		//		   ZposXpol =  9, ZnegXpol = 2, ZposYpol =  8, ZnegYpol = 4
		

		excitation(); // default constructor

		// explicitely defines the demension of the excitation plane in the mesh
		excitation(int prop_dir, int polz_dir, int xlow, int xhi, int ylow, int yhi, int zlow, int zhi);
		
		excitation(char fileName); // TBD: not implemented yet.  read excitation from file 
		~excitation(); // destuctor



		void point_set(int x, int y, int z);  //TBD
		void line_set(int xlow, int xhi, int ylow, int yhi, int zlow, int zhi);// TBD
		void plane_set(int xlow, int xhi, int ylow, int yhi, int zlow, int zhi);// TBD
		void cube_set(int xlow, int xhi, int ylow, int yhi, int zlow, int zhi); // TBD

		void point_get(int* x, int* y, int* z); // TBD
		void line_get( int* xlow, int* xhi, int* ylow, int* yhi, int* zlow, int* zhi); // TBD
		void plane_get( int* xlow, int* xhi, int* ylow, int* yhi, int* zlow, int* zhi); // TBD
		void cube_get(int* xlow, int* xhi, int* ylow, int* yhi, int* zlow, int* zhi);// TBD
		
		void freq_set(float freq);// TBD
		void delta_time_set(float dT); // TBD
		void delta_node_dist_set(float dNode); // TBD
		void excite_points_num_set(int excite_points_num);// TBD

		// public variables
		int ojb_ID; // index ID of this object
		float freq_get();// TBD
		float delta_time_get();// TBD 
		float delta_node_dist_get(); // TBD
		int excite_points_num_set();// TBD
		//float excitation_values[2000];
		float* excitation_data; // to be instantiated with " excitation_data = NEW float[n]" or malloc
		int excite_array_size;  // size of excitation waveform array

		int polorize_dir; //NONE =0, X = 1, Y = 2, Z = 3
		int propogation_dir; // NONE = 0, Yneg = 1, Zneg = 2, Xneg = 3,
		int scn_arm; // derived from polorize_dir and propogation_dir (1-12)
		int xlow, xhi, ylow, yhi, zlow, zhi;
		float freq;  // freq (Hz)
		float dT; //time step (sec)
		float dNode; // distance between nodes (m)
		
		// Gererate an excitation waveform 
		void generate_excitation_SineGaussian_array(float freq, float dNode, float height, float width, float peak_position,  int excite_array_size=2000);
		
		void excite_file_read(char fileName); //TBD: not implemented yet:  read existing excitation profile from .csv file
		void save_excitation( char* filename); // save a copy of the excitation to a .csv file (comma separated)

	private:
		
		int dFreq; // delta freq
		//int excite_points_num; 
		
		int excite_type; // NONE=0, POINT=1, HALF_SINE=2, GAUSSIAN_PULSE=3  // TBD not fully implemented
		
		// scn_arm reference: 
		//ALL = 0, XposYpol = 11, XnegYpol = 3, XposZpol = 10, XnegZpol = 6, 
		//		   YposXpol = 12, YnegXpol = 1, YposZpol =  7, YnegZpol = 5, 
		//		   ZposXpol =  9, ZnegXpol = 2, ZposYpol =  8, ZnegYpol = 4
		

	
	}; // end of excitation class



} // end of GPU_Space NAMESPACE
